@@ -4,11 +4,14 @@
 import importlib
 import logging
 import os
+import re
 import sys
-from functools import cached_property
+import weakref
 
 import json5
 
+from volkanic.compat import cached_property
+
 logger = logging.getLogger(__name__)
 
 
@@ -16,69 +19,95 @@ class Singleton(object):
     registered_instances = {}
 
     def __new__(cls, *args, **kwargs):
-        if cls not in cls.registered_instances:
+        try:
+            return cls.registered_instances[cls]
+        except KeyError:
             obj = super(Singleton, cls).__new__(cls, *args, **kwargs)
-            cls.registered_instances[cls] = obj
-        return cls.registered_instances[cls]
+            return cls.registered_instances.setdefault(cls, obj)
 
 
-def _path_join(*paths):
+class WeakSingleton(object):
+    registered_instances = weakref.WeakValueDictionary()
+
+
+class SingletonMeta(type):
+    registered_instances = {}
+
+    def __call__(cls, *args, **kwargs):
+        try:
+            return cls.registered_instances[cls]
+        except KeyError:
+            obj = super().__call__(*args, **kwargs)
+            return cls.registered_instances.setdefault(cls, obj)
+
+
+def _abs_path_join(*paths):
     path = os.path.join(*paths)
     return os.path.abspath(path)
 
 
-class GlobalInterface(Singleton):
-    # for envvar prefix and default conf paths, [a-z]+
-    primary_name = 'volcanic'
+class _GIMeta(SingletonMeta):
+    def __new__(mcs, name, bases, attrs):
+        pn = attrs.get('package_name')
+        if not pn:
+            msg = '{}.package_name is missing'.format(name)
+            raise ValueError(msg)
+        if not re.match(r'\w[\w.]*\w', pn):
+            msg = 'invalid {}.package_name: "{}"'.format(name, pn)
+            raise ValueError(msg)
+        return super().__new__(mcs, name, bases, attrs)
 
+
+class GlobalInterface(metaclass=_GIMeta):
     # for package dir, [a-z.]+
     package_name = 'volkanic'
 
-    meta_config = {
-        'src_depth': 0,
-        'filename': 'config.json5',
-    }
+    # for project dir locating (under_project_dir())
+    project_source_depth = 0
 
     # default config and log format
     default_config = {}
     default_logfmt = \
         '%(asctime)s %(levelname)s [%(process)s,%(thread)s] %(name)s %(message)s'
 
+    @cached_property
+    def project_name(self):
+        return self.package_name.replace('.', '-')
+
+    @cached_property
+    def identifier(self):
+        # for envvar prefix and default conf paths, [a-z]+
+        return self.package_name.replace('.', '_')
+
     @classmethod
     def _fmt_envvar_name(cls, name):
-        return '{}_{}'.format(cls.primary_name, name).upper()
+        return '{}_{}'.format(cls.identifier, name).upper()
 
     @classmethod
     def _get_conf_search_paths(cls):
         """
         Make sure this method can be called without arguments.
+        Override this method in your subclasses for your specific project.
         """
-        filename = cls.meta_config['filename']
-        tmpls = [
-            '/etc/{}/{}',
-            '/{}/{}',
-            '/data/local/{}/{}',
-            '/data/{}/{}',
-            os.path.expanduser('~/.{}/{}'),
+        envvar_name = cls._fmt_envvar_name('confpath')
+        pn = cls.project_name
+        return [
+            os.environ.get(envvar_name),
+            cls.under_project_dir('config.json5'),
+            cls.under_home_dir('.{}/config.json5'.format(pn)),
+            '/etc/{}/config.json5'.format(pn),
         ]
-        paths = [p.format(cls.primary_name, filename) for p in tmpls]
-        paths += [cls.under_project_dir(filename)]
-        return paths
 
     @classmethod
     def _locate_conf(cls):
         """
         Returns: (str) absolute path to config file
         """
-        envvar_name = cls._fmt_envvar_name('confpath')
-        try:
-            return os.environ[envvar_name]
-        except KeyError:
-            pass
         for path in cls._get_conf_search_paths():
-            path = os.path.abspath(path)
+            if not path:
+                continue
             if os.path.exists(path):
-                return path
+                return os.path.abspath(path)
 
     @staticmethod
     def _parse_conf(path: str):
@@ -88,7 +117,8 @@ class GlobalInterface(Singleton):
     def conf(self) -> dict:
         path = self._locate_conf()
         if path:
-            print('GlobalInterface.conf, path', path, file=sys.stderr)
+            msg = '{}.conf, path'.format(self.__class__.__name__)
+            print(msg, path, file=sys.stderr)
             user_config = self._parse_conf(path)
         else:
             user_config = {}
@@ -96,10 +126,13 @@ class GlobalInterface(Singleton):
         config.update(user_config)
         return config
 
-    def under_data_dir(self, *paths) -> str:
-        dirpath = self.conf['data_dir']
-        path = os.path.join(dirpath, *paths)
-        return os.path.abspath(path)
+    @staticmethod
+    def under_home_dir(*paths):
+        if sys.platform == 'win32':
+            homedir = os.environ["HOMEPATH"]
+        else:
+            homedir = os.path.expanduser('~')
+        return _abs_path_join(homedir, *paths)
 
     @classmethod
     def under_package_dir(cls, *paths) -> str:
@@ -107,19 +140,17 @@ class GlobalInterface(Singleton):
         pkg_dir = os.path.split(mod.__file__)[0]
         if not paths:
             return pkg_dir
-        path = os.path.join(pkg_dir, *paths)
-        return os.path.abspath(path)
-
-    def under_resources_dir(self, *paths):
-        dirpath = self.conf['resources_dir']
-        return os.path.join(dirpath, *paths)
+        return _abs_path_join(pkg_dir, *paths)
 
     @classmethod
     def under_project_dir(cls, *paths):
-        n = cls.meta_config.get('src_depth', 0)
-        n += len(cls.package_name.split())
+        pkg_dir = cls.under_package_dir()
+        if re.search(r'[/\\](site|dist)-packages[/\\]', pkg_dir):
+            return
+        n = cls.project_source_depth
+        n += len(cls.package_name.split('.'))
         paths = ['..'] * n + list(paths)
-        return cls.under_package_dir(*paths)
+        return os.path.join(pkg_dir, *paths)
 
     @cached_property
     def jinja2_env(self):
